#
# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019. Built in San Francisco.
#
# This software is distributed under commercial non-GPL license for personal, educational,
# corporate or any other use. The software as a whole or any parts of that are prohibited
# for distribution and/or use without obtaining license from 3D Control Systems, Inc.
#
# If you do not have the license to use this software, please delete all software files
# immediately and contact sales to obtain the license: sales@3dprinteros.com.
# If you are unsure about the licensing please contact directly our sales: sales@3dprinteros.com.

import sys
import time
import signal
import base64

import log
import paths
import config
import user_login
import http_client


class Camera:

    MAX_CAMERA_INDEX = 10
    FAILS_BEFORE_REINIT = 10
    X_RESOLUTION = 640.0
    Y_RESOLUTION = 480.0

    @log.log_exception
    def __init__(self):
        self.logger = log.create_logger('camera')
        self.stop_flag = False
        paths.init_path_to_libs()
        import numpy as np
        import cv2 as cv2
        self.np = np
        self.cv2 = cv2
        signal.signal(signal.SIGINT, self.intercept_signal)
        signal.signal(signal.SIGTERM, self.intercept_signal)
        if len(sys.argv) > 2:
            self.user_token = sys.argv[1]
            mac = sys.argv[2]
        else:
            ul = user_login.UserLogin(self)
            ul.wait()
            self.user_token = ul.user_token
            mac = None
        self.image_extension = config.get_settings()["camera"]["img_ext"]
        self.image_quality = config.get_settings()["camera"]["img_qual"]
        self.hardware_resize = config.get_settings()["camera"]["hardware_resize"]
        self.min_loop_time = config.get_settings()["camera"]["min_loop_time"]
        self.search_cameras()
        self.http_client = http_client.HTTPClient(self, keep_connection_flag=True)
        if mac:
            self.http_client.mac = mac #we need to use MAC from client to ensure that it's not changed on camera restart
        self.main_loop()

    def intercept_signal(self, signal_code, frame):
        self.logger.info("SIGINT or SIGTERM received. Closing Camera Module...")
        self.close()

    def close(self):
        self.stop_flag = True

    def search_cameras(self):
        self.captures = []
        self.fails = []
        self.resized = []
        for index in range(0, self.MAX_CAMERA_INDEX):
            if self.stop_flag:
                break
            self.logger.debug("Probing for camera N%d..." % index)
            capture = self.cv2.VideoCapture(index)
            if capture.isOpened():
                self.init_capture(capture)
                self.logger.info("...got camera at index %d" % index)
            else:
                del(capture)
        self.logger.info("Got %d operational cameras" % len(self.captures))

    def init_capture(self, capture):
        self.resized.append(self.set_resolution(capture))
        self.captures.append(capture)
        self.fails.append(0)

    def set_resolution(self, cap):
        if self.hardware_resize:
            try:
                x = cap.get(self.cv2.cv.CV_CAP_PROP_FRAME_WIDTH)
                y = cap.get(self.cv2.cv.CV_CAP_PROP_FRAME_HEIGHT)
            except AttributeError:
                return False
            if x == self.X_RESOLUTION and y == self.Y_RESOLUTION:
                return True
            # protection against setting wrong parameters(some cameras have different params on this indexes)
            elif x > 100 and y > 100:
                try:
                    result = cap.set(self.cv2.cv.CV_CAP_PROP_FRAME_WIDTH, self.X_RESOLUTION) and \
                             cap.set(self.cv2.cv.CV_CAP_PROP_FRAME_HEIGHT, self.Y_RESOLUTION)
                except:
                    result = False
                return result

    def make_shot(self, capture):
        #self.logger.debug("Capturing frame from " + str(capture))
        index = self.captures.index(capture)
        try:
            state, frame = capture.read()
        except Exception as e:
            self.logger.error("Error while capturing frame: " + str(e))
        else:
            if state and frame.any():
                self.fails[index] = 0
            else:
                self.fails[index] += 1
                return
            if not self.resized[index]: # resize using software
                try:
                    frame = self.cv2.resize(frame, (self.X_RESOLUTION, self.Y_RESOLUTION))
                except TypeError:
                    # some opencv version accept integer args here
                    self.X_RESOLUTION = int(self.X_RESOLUTION)
                    self.Y_RESOLUTION = int(self.Y_RESOLUTION)
                    self.fails[index] += 1
                    return self.make_shot(capture)
                except Exception as e:
                    self.resized[index] = True
                    self.logger.warning("Error while software resize of frame: " + str(e))
            encode_param = [int(self.cv2.IMWRITE_JPEG_QUALITY), self.image_quality]
            try:
                result, encoded_frame = self.cv2.imencode(self.image_extension, frame, encode_param)
            except Exception as e:
                self.logger.warning('Failed to encode camera frame: ' + str(e))
                result, encoded_frame = None, None
            if state and result:
                data = self.np.array(encoded_frame)
                string_data = data.tostring()
                #self.logger.debug("Successfully captured and encoded from" + str(capture))
                return string_data

    def send_frame(self, number, frame):
        frame = base64.b64encode(str(frame))
        number = number + 1
        message = self.user_token, number, "Camera" + str(number), frame
        #self.logger.debug("Camera %d sending frame to server..." % number)
        answer = self.http_client.pack_and_send('camera', *message)
        if not answer:
            self.logger.debug("Camera %d can't send frame to server - HTTP error" % number)

    def main_loop(self):
        while not self.stop_flag:
            frame_start_time = time.time()
            reinit_needed = False
            for number, capture in enumerate(self.captures):
                if self.fails[number] > self.FAILS_BEFORE_REINIT:
                    reinit_needed = True
                frame = self.make_shot(capture)
                if frame:
                    #self.logger.debug("Got frame from camera N" + str(number))
                    self.send_frame(number, frame)
                else:
                    self.logger.warning("No frame from camera N" + str(number))
            while time.time() < frame_start_time + 0.001 + self.min_loop_time * len(self.captures):
                time.sleep(0.01)
            if reinit_needed:
                self.logger.warning("Error of camera %d. Starting cameras reinitialisation..." % number)
                self.close_captures()
                self.search_cameras()
                self.logger.warning("...done reinitialising cameras.")
        self.close_captures()
        self.http_client.close()
        sys.exit(0)

    def close_captures(self):
        for capture in self.captures:
            capture.release()
            self.logger.info("Closed camera capture " + str(capture))
            #del(capture)

    def register_error(self, code, message, is_blocking=False):
        self.logger.warning("Error N%d. %s" % (code, message))


if __name__ == '__main__':
    Camera()